/************************************************************************
*
* lnk980 - Links objects from asm980
*
* Changes:
*   06/29/11   DGP   Original. Hacked from lnk990.
*   08/04/11   DGP   Added 980 card object format. 
*	
************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <time.h>

#include "lnkdef.h"

FILE *lstfd = NULL;
int listmode = FALSE;
int pc; /* the linker pc */
int absolute = FALSE;
int partiallink = FALSE;
int symbolcount = 0;
int errcount = 0;
int pgmlength = 0;
int absentry = -1;
int relentry = -1;
int linecnt = MAXLINE;
int modcount = 0;
int undefs = FALSE;
int muldefs = FALSE;
int cardoutmode = FALSE;
int cardinmode = FALSE;
int curmod = 0;
int extndx = 0;
char errline[120];
char inbuf[MAXLINE];
char idtbuf[IDTSIZE+2];
struct tm *timeblk;

SymNode *symbols[MAXSYMBOLS];
Module modules[MAXMODULES];

uint16 memory[MEMSIZE];
Memory memctl[MEMSIZE];

static int pagenum = 0;
static char datebuf[48];

/***********************************************************************
* printheader - Print header on listing.
***********************************************************************/

void
printheader (FILE *lstfd)
{
   if (linecnt >= LINESPAGE)
   {
      linecnt = 0;
      if (pagenum) fputc ('\f', lstfd);
      fprintf (lstfd, H1FORMAT,
	       VERSION,
	       idtbuf,
	       datebuf,
	       ++pagenum);
   }
}

/***********************************************************************
* printsymbols - Print the symbol table.
***********************************************************************/

static void
printsymbols (FILE *lstfd)
{
   SymNode *sym;
   int i, j;
   int longformat;
   int maxsymlen;
   int symperline;
   char type;
   char type1;
   char format[40];

   longformat = FALSE;
   maxsymlen = 8;

   fprintf (lstfd, "\n                        D E F I N I T I O N S \n\n");
   linecnt += 3;

   symperline = 80 / (12 + maxsymlen);

   j = 0;
   for (i = 0; i < symbolcount; i++)
   {
      sym = symbols[i];
      printheader (lstfd);

      if (partiallink)
      {
	 if (sym->flags & GLOBAL) type = 'D';
	 else if (sym->flags & EXTERNAL) type = 'R';
      }
      else if (sym->flags & RELOCATABLE) type = '\'';
      else type = ' ';

      if (sym->flags & MULDEF)
      {
	 muldefs = TRUE;
	 type1 = 'M';
      }
      else if (sym->flags & UNDEF && sym->value != 0)
      {
	 if (!partiallink && !(sym->flags & SREF))
	 {
	    errcount++;
	    undefs = TRUE;
	    type1 = 'U';
	 }
      }
      else type1 = ' ';

      if (longformat)
      {
	 fprintf (lstfd, format,
		  sym->symbol, sym->value,
		  type, type1, sym->modnum);
	 j++;
      }
      else
      {
	 fprintf (lstfd, SYMFORMAT,
		  sym->symbol, sym->value,
		  type, type1, sym->modnum);
	 j++;
      }
      if (j == symperline)
      {
	 fprintf (lstfd, "\n");
	 linecnt++;
	 printheader(lstfd);
	 j = 0;
      }
   }
   fprintf (lstfd, "\n");
}

/***********************************************************************
* scansymbol - Scan the symbol table for undefined symbols.
***********************************************************************/

static void
scansymbols (void)
{
   SymNode *sym;
   int i;

   for (i = 0; i < symbolcount; i++)
   {
      sym = symbols[i];
      if (!partiallink && sym->flags & UNDEF)
      {
	 if (!partiallink && !(sym->flags & SREF))
	 {
	    fprintf (stderr, "Undefined symbol: %s\n", sym->symbol);
	    undefs = TRUE;
	    errcount++;
	 }
      }
   }
}

/***********************************************************************
* Main procedure
***********************************************************************/

int
main (int argc, char **argv)
{
   FILE *infd = NULL;
   FILE *outfd = NULL;
   char *infile[MAXFILES];
   char *libs[MAXFILES];
   char *libdirs[MAXFILES];
   char *outfile = NULL;
   char *lstfile = NULL;
   char *bp;
   int i;
   int listMode = FALSE;
   int incnt = 0;
   int libcnt = 0;
   int libdircnt = 0;
   int status = 0;
   time_t curtime;
   char temp[256];
  
#ifdef DEBUG
   printf ("lnk980: Entered:\n");
   printf ("args:\n");
   for (i = 1; i < argc; i++)
      printf ("   arg[%2d] = %s\n", i, argv[i]);
#endif

   /*
   ** Clear files.
   */

   for (i = 0; i < MAXFILES; i++)
   {
      infile[i] = NULL;
      libs[i] = NULL;
      libdirs[i] = NULL;
   }

   /*
   ** Clear symboltable
   */

   for (i = 0; i < MAXSYMBOLS; i++)
      symbols[i] = NULL;

   /*
   ** Scan off the the args.
   */

   idtbuf[0] = '\0';

   for (i = 1; i < argc; i++)
   {
      bp = argv[i];

      if (*bp == '-')
      {
	 bp++;
         while (*bp) switch (*bp)
         {
	 case 'c':
	    cardoutmode = TRUE;
	    bp++;
	    break;

	 case 'i':
	    i++;
	    if (i >= argc) goto USAGE;
	    if (strlen(argv[i]) >= IDTSIZE) argv[i][IDTSIZE] = '\0';
	    strcpy (idtbuf, argv[i]);
	    bp++;
	    break;

	 case 'L':
	    if (libdircnt >= MAXFILES)
	    {
	       fprintf (stderr, "Maximum library directories exceeded: %d\n",
	       		MAXFILES);
	       exit (ABORT);
	    }
	    libdirs[libdircnt++] = bp+1;
	    while (*bp) bp++;
	    break;

	 case 'l':
	    if (libcnt >= MAXFILES)
	    {
	       fprintf (stderr, "Maximum libraries exceeded: %d\n", MAXFILES);
	       exit (ABORT);
	    }
	    libs[libcnt++] = bp+1;
	    while (*bp) bp++;
	    break;

	 case 'M':
	    listMode = TRUE;
	    bp++;
	    break;

	 case 'm':
	    i++;
	    if (i >= argc) goto USAGE;
	    lstfile = argv[i];
	    listmode = TRUE;
	    bp++;
	    break;

         case 'o':
            i++;
	    if (i >= argc) goto USAGE;
            outfile = argv[i];
	    bp++;
            break;

	 case 'p':
	    partiallink = TRUE;
	    bp++;
	    break;

         default:
      USAGE:
	    printf ("lnk980 - version %s\n", VERSION);
	    printf ("usage: lnk980 [-options] -o file.bin infile.bin...\n");
            printf (" options:\n");
	    printf ("    -c           - Card object mode\n");
            printf ("    -i IDT       - Define IDT name for link\n");
            printf ("    -llib        - Library name\n");
            printf ("    -Llibdir     - Library directory\n");
	    printf ("    -m listfile  - Generate map listing to listfile\n");
	    printf ("    -o objfile   - Linked object to objfile\n");
	    printf ("    -p           - Partial linking\n");
	    return (ABORT);
         }
      }

      else
      {
	 if (incnt >= MAXFILES)
	 {
	    fprintf (stderr, "lnk980: Maximum input files exceeded: %d\n",
		     MAXFILES);
	    exit (ABORT);
	 }
         infile[incnt++] = argv[i];
      }

   }

   if (!infile[0] && !outfile) goto USAGE;
   if (partiallink && cardoutmode)
   {
      fprintf (stderr,
	       "lnk980: Partial linking not supported with card object\n");
      exit (ABORT);
   }

#ifdef DEBUG
   printf (" outfile = %s\n", outfile);
   for (i = 0; i < incnt; i++)
      printf (" infile[%d] = %s\n", i, infile[i]);
   for (i = 0; i < libcnt; i++)
      printf (" libs[%d] = %s\n", i, libs[i]);
   for (i = 0; i < libdircnt; i++)
      printf (" libdirs[%d] = %s\n", i, libdirs[i]);
   if (listmode)
      printf (" lstfile = %s\n", lstfile);
#endif

   /*
   ** Open the files.
   */

   if ((outfd = fopen (outfile, "w")) == NULL)
   {
      perror ("lnk980: Can't open output file");
      exit (ABORT);
   }
   if (listMode)
   {
      listmode = TRUE;
      sprintf (temp, "%s.map", outfile);
      lstfile = temp;
   }

   if (listmode)
   {
      if ((lstfd = fopen (lstfile, "w")) == NULL)
      {
	 perror ("lnk980: Can't open listing file");
	 exit (ABORT);
      }

   }

   /*
   ** Get current date/time.
   */

   curtime = time(NULL);
   timeblk = localtime (&curtime);
   strcpy (datebuf, ctime(&curtime));
   *strchr (datebuf, '\n') = '\0';

   /*
   ** If IDT not specified, use output file name
   */

   if (!idtbuf[0])
   {
      char *sp, *ep;
      char temp[256];

      strcpy (temp, outfile);
      if ((sp = strrchr (temp, '/')) == NULL)
         sp = temp;
      else
         sp++;
      if ((ep = strchr (sp, '.')) != NULL)
         *ep = '\0';
      if (strlen(sp) >= IDTSIZE) sp[IDTSIZE] = '\0';
      strcpy (idtbuf, sp);
#ifdef DEBUG
      printf (" idtbuf = %s\n", idtbuf);
#endif
   }
   for (i = 0; idtbuf[i]; i++)
   {
      if (islower (idtbuf[i])) 
         idtbuf[i] = toupper (idtbuf[i]);
   }

   /*
   ** Clear memory.
   */

   memset (&memory, '\0', sizeof(memory));
   for (i = 0; i < MEMSIZE; i++)
   {
      memset (&memctl[i], '\0', sizeof (Memory));
      memctl[i].tag = ABSDATA_TAG;
   }

   /*
   ** Load the objects for pass 1.
   */

   pc = 0;
   curmod = 0;
   for (i = 0; i < incnt; i++)
   {
      if ((infd = fopen (infile[i], "r")) == NULL)
      {
	 sprintf (inbuf, "lnk980: Can't open input file %s", infile[i]);
	 perror (inbuf);
	 exit (ABORT);
      }

      pc = lnkloader (infd, pc, 1, infile[i], FALSE);

      fclose (infd);
   }

   /*
   ** Clear memory.
   */

   memset (&memory, '\0', sizeof(memory));
   for (i = 0; i < MEMSIZE; i++)
   {
      memset (&memctl[i], '\0', sizeof (Memory));
      memctl[i].tag = ABSDATA_TAG;
   }

   /*
   ** reLoad the objects for pass 2.
   */

   pc = 0;
   curmod = 0;
   for (i = 0; i < incnt; i++)
   {

      if ((infd = fopen (infile[i], "r")) == NULL)
      {
	 sprintf (inbuf, "lnk980: Can't open input file %s", infile[i]);
	 perror (inbuf);
	 exit (ABORT);
      }

      pc = lnkloader (infd, pc, 2, infile[i], FALSE);

      fclose (infd);
   }

   /*
   ** If libraries, search them.
   */

   if (libcnt)
   {
      curmod = modcount;
      for (i = 0; i < libcnt; i++)
      {
	 int j;
	 char libfile[1024];

	 if (libdircnt)
	 {
	    for (j = 0; j < libdircnt; j++)
	    {
	       sprintf (libfile, "%s/lib%s.a", libdirs[j], libs[i]);

	       if ((infd = fopen (libfile, "r")) != NULL)
	       {
#ifdef DEBUG
		  printf (" Library: %s\n", libfile);
#endif
		  sprintf (libfile, "lib%s", libs[i]);
		  status = lnklibrary (infd, pc, 3, libfile);
		  fclose (infd);
	       }
	    }
	 }
	 else
	 {
	    sprintf (libfile, "lib%s.a", libs[i]);

	    if ((infd = fopen (libfile, "r")) != NULL)
	    {
#ifdef DEBUG
	       printf (" Library: %s\n", libfile);
#endif
	       sprintf (libfile, "lib%s", libs[i]);
	       status = lnklibrary (infd, pc, 3, libfile);
	       fclose (infd);
	    }
	 }
      }
   }

   /*
   ** Go back and fill in all defined EXTVALUE records.
   */

   fillinextndx();

   if (listmode)
   {

      /*
      ** Print the modules
      */

      printheader (lstfd);
      fprintf (lstfd,
	 "MODULE   NO  ORIGIN  LENGTH    DATE      TIME    CREATOR   FILE\n\n");
      linecnt += 2;
      for (i = 0; i < modcount; i++)
      {
	 printheader (lstfd);
         fprintf (lstfd, MODFORMAT, modules[i].name, i+1, modules[i].origin,
		  modules[i].length, modules[i].date, modules[i].time,
		  modules[i].creator, modules[i].objfile);
	 linecnt ++;
	 printheader(lstfd);
      }

      /*
      ** Print the symbol table
      */

      printsymbols (lstfd);

      printheader (lstfd);
      fprintf (lstfd, "\n");
      linecnt ++;
      printheader(lstfd);

      if (undefs)
      {
	 fprintf (lstfd, "** There are undefined symbols - U\n");
	 linecnt ++;
	 printheader(lstfd);
      }
      if (muldefs)
      {
	 fprintf (lstfd, "** There are multiply defined symbols - M\n");
	 linecnt ++;
	 printheader(lstfd);
      }

      fprintf (lstfd, "\n%04X Length of program \n", pc);
      linecnt ++;
      printheader(lstfd);
      if (relentry >= 0)
      {
	 fprintf (lstfd, "%04X Program entry \n", relentry);
	 linecnt ++;
	 printheader(lstfd);
      }
      fprintf (lstfd, "%d Errors\n", errcount);
   }
   else
      scansymbols();

   if (errcount)
   {
      printf ("lnk980: %d errors\n", errcount);
      status = 1;
   }

   /*
   ** Punch out linked object
   */

   if (!errcount)
      lnkpunch (outfd, partiallink);

   /*
   ** Close the files
   */

   fclose (outfd);
   if (listmode)
      fclose (lstfd);

   return (status == 0 ? NORMAL : ABORT);
}
