/***********************************************************************
*
* lnkdef.h - Linker header for the TI 980 computer.
*
* Changes:
*   06/29/11   DGP   Original. Hacked from lnk990.
*   08/04/11   DGP   Added 980 card object format.
*
***********************************************************************/

#include <stdio.h>

/*
** Definitions
*/

#define NORMAL 0
#define ABORT  12

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#define VERSION "1.1.1"

#define EOFSYM ':'

#define MAXLINE 256
#define LINESPAGE 60
#define MAXFILES 1000
#define IDTSIZE 6

#define MAXMODULES 1000
#define MAXSYMLEN 6
#define MAXSYMBOLS 5000

#define MEMSIZE 32768

#define TEMPSPEC "lnkXXXXXX"

/*
** Object output formats
*/

#define OBJFORMAT "%c%04X"
#define IDTFORMAT "%c%4.4X%-6.6s"
#define REFFORMAT "%c%4.4X%-6.6s"
#define DEFFORMAT "%c%4.4X%-6.6s"
#define SEQFORMAT "%4.4d\n"

#define CARDIDTFORMAT "%-6.6s%1.1d%04X%04X%3.3d%3.3d"
#define CARDREFFORMAT "%c%-6.6s"
#define CARDDEFFORMAT "%c%-6.6s%04X"

/*
** Listing formats
*/

#define H1FORMAT "LNK980 %-8.8s  %-24.24s  %s Page %04d\n\n"

#define MODFORMAT "%-8.8s %2d   %04X    %04X   %s  %s  %s  %s\n"
#define SYMFORMAT "%-8.8s %4.4X%c%c %2d  "

/*
** Data type definitions
*/

#define int8            char
#define int16           short
#define int32           int
typedef int             t_stat;                         /* status */
typedef int             t_bool;                         /* boolean */
typedef unsigned int8   uint8;
typedef unsigned int16  uint16;
typedef unsigned int32  uint32, t_addr;                 /* address */

#if defined (WIN32)                                     /* Windows */
#define t_int64 __int64
#elif defined (__ALPHA) && defined (VMS)                /* Alpha VMS */
#define t_int64 __int64
#elif defined (__ALPHA) && defined (__unix__)           /* Alpha UNIX */
#define t_int64 long
#else                                                   /* default GCC */
#define t_int64 long long
#endif                                                  /* end OS's */
typedef unsigned t_int64        t_uint64, t_value;      /* value */
typedef t_int64                 t_svalue;               /* signed value */

#define MSB 0
#define LSB 1

/*
** Memory
*/

typedef struct
{
   char tag;
   int  flags;
   int	value;
} Memory;

#define GETMEM(m) memory[m]

#define PUTMEM(m,v) memory[m]=v

/*
** Symbol table
*/

typedef struct EXTndx_t
{
    struct EXTndx_t *next;
    int location;
} EXTndxlist;

typedef struct
{
   EXTndxlist *extlist;
   char symbol[MAXSYMLEN+2];
   char module[MAXSYMLEN+2];
   int flags;
   int value;
   int modnum;
   int extndx;
} SymNode;

#define RELOCATABLE 0x000000001
#define EXTERNAL    0x000000002
#define GLOBAL      0x000000004
#define MULDEF      0x000000010
#define UNDEF       0x000000020
#define ABSDEF      0x000000100
#define SREF        0x000000200
#define LOAD        0x000000400

/*
** Module table
*/

typedef struct 
{
   char name[IDTSIZE+2];
   char date[12];
   char time[12];
   char creator[12];
   char objfile[256];
   int origin;
   int length;
} Module;

/*
** Object tags
*/

#define IDT_TAG		'0'
#define ABSENTRY_TAG	'1'
#define RELENTRY_TAG	'2'
#define RELEXTRN_TAG	'3'
#define ABSEXTRN_TAG	'4'
#define RELGLOBAL_TAG	'5'
#define ABSGLOBAL_TAG	'6'
#define CKSUM_TAG	'7'
#define NOCKSUM_TAG	'8'
#define ABSORG_TAG	'9'
#define RELORG_TAG	'A'
#define ABSDATA_TAG	'B'
#define RELDATA_TAG	'C'
#define EXTVALUE_TAG	'D'
#define RELBYTE_TAG	'E'
#define EOR_TAG		'F'

#define CHARSPERREC	66  /* Chars per object record */
#define WORDLEN		4   /* Word length */
#define WORDTAGLEN	5   /* Word + Object Tag length */
#define BINWORDTAGLEN	3   /* Word + Binary Object Tag length */
#define EXTRNLEN	11  /* Tag + SYMBOL + addr */
#define GLOBALLEN	11  /* Tag + SYMBOL + addr */
#define IDTLEN		11  /* Tag + IDT + length */
#define BINIDTLEN	9   /* Tag + IDT + binary length */
#define IDTCOLUMN	72  /* Where to put the IDT */
#define SEQUENCENUM	76  /* Where to put the sequence */

#define TIMEOFFSET 17
#define DATEOFFSET 27
#define CREATOROFFSET 38


/*
** External functions.
*/

extern int lnkloader (FILE *, int, int, char *, int);
extern uint16 getnumfield (unsigned char *, int);
extern void fillinextndx (void);

extern int lnklibrary (FILE *, int, int, char *);

extern int lnkpunch (FILE *, int);

extern void printheader (FILE *);
extern SymNode *symlookup (char *, char *, int);
