/************************************************************************
*
* lnkloader - Loads objects from asm980 for linking.
*
* Changes:
*   06/29/11   DGP   Original. Hacked from lnk990.
*
************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <ctype.h>
#include <errno.h>
#include <memory.h>

#include "lnkdef.h"

extern int absentry;
extern int relentry;
extern int modcount;
extern int symbolcount;
extern int extndx;
extern int curmod;
extern int cardinmode;

extern uint16 memory[MEMSIZE];
extern Memory memctl[MEMSIZE];
extern SymNode *symbols[MAXSYMBOLS];
extern char idtbuf[IDTSIZE+2];
extern Module modules[MAXMODULES];


/************************************************************************
* getnumfield - Gets the next numeric field.
************************************************************************/

uint16
getnumfield (unsigned char *op, int binarymode)
{
   uint16 wdata;
   char item[6];

   if (binarymode)
   {
      wdata = (*op << 8) | *(op+1);
   }
   else
   {
      int wd;

      strncpy (item, (char *)op, 4);
      item[4] = '\0';
      sscanf (item, "%4X", &wd);
      wdata = wd & 0xFFFF;
   }
   return (wdata);
}

/************************************************************************
* processdef - Process DEF.
************************************************************************/

static void
processdef (char *item, char *module, int pass, int relo,
	    int flags, uint16 wdata, int libload)
{
   SymNode *s;
   char *bp;

   if (pass == 1)
   {
      bp = item;
      for (; *bp; bp++) if (isspace(*bp)) *bp = '\0';
#ifdef DEBUGLOADER
      printf (", %c%s = %s, libload = %s",
	    relo ? 'R' : 'A', "DEF",
	    item, libload ? "TRUE" : "FALSE");
#endif
      if ((s = symlookup (item, module, FALSE)) == NULL)
      {
	 if ((s = symlookup (item, module, TRUE)) == NULL)
	 {
	    fprintf (stderr, "lnk980: Internal error\n");
	    exit (ABORT);
	 }
	 s->flags = relo ? RELOCATABLE : ABSDEF;
	 s->flags |= GLOBAL | flags;
	 s->value = wdata;
	 if (libload)
	    s->modnum = curmod + 1;
#ifdef DEBUGLOADER
	 printf ("\n      flags = %04X, value = %04X",
	       s->flags, s->value);
#endif
      }
      else
      {
	 if (libload)
	 {
	    s->modnum = curmod + 1;
	    if (s->flags & UNDEF)
	    {
	       int refaddr;

#ifdef DEBUGLOADER
	       printf ("\n      fill: flags = %04X, value = %04X",
		     s->flags, s->value);
#endif
	       refaddr = s->value;
	       while (refaddr)
	       {
		  int k;
		  k = GETMEM (refaddr);
		  PUTMEM (refaddr, wdata);
		  memctl[refaddr].tag = relo ? RELDATA_TAG : ABSDATA_TAG;
		  refaddr = k;
#ifdef DEBUGLOADER
		  printf ("\n          refaddr = %04X", refaddr);
#endif
	       }
	       s->flags = relo ? RELOCATABLE : ABSDEF;
	       s->flags |= GLOBAL | flags;
	       s->value = wdata;
	    }
	    else
	       s->flags |= MULDEF;
	 }
	 else
	    s->flags |= MULDEF;
      }
   }

}

/************************************************************************
* processref - Process REF.
************************************************************************/

static void
processref (char *item, char *module, int pass, int relo,
	    int flags, uint16 wdata)
{
   SymNode *s;
   char *bp;

   if (pass == 2)
   {
      bp = item;
      for (; *bp; bp++) if (isspace(*bp)) *bp = '\0';
#ifdef DEBUGLOADER
      printf (", %c%s = %s",
	    relo ? 'R' : 'A', "REF",
	    item);
#endif

      if ((s = symlookup (item, module, FALSE)) == NULL)
      {
	 if ((s = symlookup (item, module, TRUE)) == NULL)
	 {
	    fprintf (stderr, "lnk980: Internal error\n");
	    exit (ABORT);
	 }
#ifdef DEBUGLOADER
         printf (", module = %d", curmod);
#endif
	 s->flags = relo ? RELOCATABLE : 0;
	 s->flags |= EXTERNAL | UNDEF | flags;
	 s->value = wdata;
	 s->extndx = extndx++;
	 s->modnum = curmod;
      }
      else
      {
	 int refaddr;

         if (s->extndx < 0)
	    s->extndx = extndx++;
	 if (relo && !(s->flags & ABSDEF))
	    s->flags |= RELOCATABLE;
	 refaddr = wdata;
	 if (flags & LOAD)
	 {
	    if (s->flags & SREF)
	       s->flags &= ~(LOAD|SREF);
	    return;
	 }
#ifdef DEBUGLOADER
	 printf ("\n      flags = %04X, value = %04X",
		 s->flags, s->value);
#endif

	 if (!cardinmode)
	 {
	    /*
	    ** If defined, fill in back chain
	    */

#ifdef DEBUGLOADER
	    printf ("\n          refaddr = %04X", refaddr);
#endif
	    if (!(s->flags & UNDEF))
	    {
	       while (refaddr)
	       {
		  int k;

		  k = GETMEM (refaddr);
		  PUTMEM (refaddr, s->value);
		  memctl[refaddr].tag =
		       (s->flags & RELOCATABLE) ? RELDATA_TAG : ABSDATA_TAG;
		  refaddr = k;
#ifdef DEBUGLOADER
		  printf ("\n          refaddr = %04X", refaddr);
#endif
	       }
	    }

	    /*
	    ** If undefined, connect back chain
	    */

	    else
	    {
	       while (refaddr)
	       {
		  int k;

		  k = GETMEM (refaddr);
		  if (k == 0)
		  {
		     memctl[refaddr].tag = RELDATA_TAG;
		     PUTMEM(refaddr, s->value);
		     s->value = wdata;
		  }
		  refaddr = k;
	       }
	    }
	 }
      }
   }
}

/************************************************************************
* fillinextndx - Process EXTVALUE back fills.
************************************************************************/

void
fillinextndx (void)
{
   int i;

#ifdef DEBUGLOADER
   printf ("\n\nfillinextndx: entered\n");
#endif

   for (i = 0; i < symbolcount; i++)
   {
      SymNode *s;

      s = symbols[i];
      if (s->extlist)
      {
	 EXTndxlist *ref;
	 int curraddr;

	 ref = s->extlist;
	 while (ref)
	 {
	    curraddr = ref->location;
	    if (s->flags & UNDEF)
	    {
	       memctl[curraddr].tag = ABSDATA_TAG;
	       PUTMEM (curraddr, 0);
#ifdef DEBUGLOADER
	       printf (
		     "   sym = %s-U, flags = %04X, addr = %04X, value = %04X\n",
		     s->symbol, s->flags, curraddr, GETMEM(curraddr));
#endif
	    }
	    else
	    {
	       memctl[curraddr].tag =
		      (s->flags & RELOCATABLE) ? RELDATA_TAG : ABSDATA_TAG;
	       PUTMEM (curraddr, memctl[curraddr].value + s->value);
#ifdef DEBUGLOADER
	       printf (
	 "   sym = %s, flags = %04X, addr = %04X, value = %04X (%04X + %04X)\n",
		     s->symbol, s->flags, curraddr, GETMEM(curraddr),
		     memctl[curraddr].value, s->value);
#endif
	    }
	    ref = ref->next;
	 }
      }
   }
}

/************************************************************************
* processextndx - Process EXTVALUE tag.
************************************************************************/

static void
processextndx (int curraddr, int seg)
{
   int i;
   uint16 wdata;

   wdata = GETMEM (curraddr);
#ifdef DEBUGLOADER
   printf ("\n      processextndx: curraddr = %04X, wdata = %04X",
           curraddr, wdata);
#endif

   for (i = 0; i < symbolcount; i++)
   {
      SymNode *s;

      s = symbols[i];
      if (wdata == s->extndx)
      {
	 if (s->flags & UNDEF)
	 {
	    EXTndxlist *new;

	    if ((new = malloc (sizeof (EXTndxlist))) == NULL)
	    {
	       fprintf (stderr, "lnk980: Unable to allocate memory\n");
	       exit (ABORT);
	    }
#ifdef DEBUGLOADER
	    printf (
		 "\n       sym = %s-U, flags = %04X, value = %04X, addr = %04X",
		  s->symbol, s->flags, GETMEM(curraddr), curraddr);
#endif
	    new->next = s->extlist;
	    new->location = curraddr;
	    s->extlist = new;
	 }
	 else
	 {
	    memctl[curraddr].tag =
		   (s->flags & RELOCATABLE) ? RELDATA_TAG : ABSDATA_TAG;
	    PUTMEM (curraddr, memctl[curraddr].value + s->value);
#ifdef DEBUGLOADER
	    printf (
		"\n       sym-F = %s, flags = %04X, value = %04X (%04X + %04X)",
		  s->symbol, s->flags, GETMEM(curraddr),
		  memctl[curraddr].value, s->value);
#endif
	 }
	 break;
      }
   }
}

/***********************************************************************
* loadcardfile - Load a card mode object file.
***********************************************************************/

static int
loadcardfile (FILE *fd, int loadpt, int pass, char *file, int libload)
{
   int idtrec = TRUE;
   int curraddr;
   int reloprog;
   int reclen;
   int refcnt, defcnt;
   uint16 reloword;
   uint16 pseglen;
   uint8 inbuf[82];
   char module[MAXSYMLEN+2];

#ifdef DEBUGLOADER
   printf ("\n\nloadcardfile: loadpt = %04X, pass = %d, file = %s\n",
	   loadpt, pass, file);
#endif

   curraddr = loadpt;
   defcnt = refcnt = 0;
   pseglen = 0;
   while (fgets ((char *)inbuf, 82, fd) != NULL)
   {
      uint8 *op = (uint8 *)inbuf;
      uint16 wdata;
      char item[80];

      reclen = strlen ((char *)inbuf);

      /*
      ** The first "card" is the IDT record.
      */

      if (idtrec)
      {
	 int i;
	 char *tp;

         idtrec = FALSE;

	 /* Sanity check */

	 tp = (char *)op + IDTSIZE;
	 if (*tp != '0' && *tp != '1')
	    goto BAD_OBJECT;
	 tp++;
	 for (i = 0; i < 8; i++)
	    if (!isxdigit(*tp++))
	       goto BAD_OBJECT;
	 for (i = 0; i < 3; i++)
	 {
	    if (!isdigit(*tp))
	       goto BAD_OBJECT;
	    defcnt = (defcnt * 10) + (*tp - '0');
	    tp++;
	 }
	 for (i = 0; i < 3; i++)
	 {
	    if (!isdigit(*tp))
	       goto BAD_OBJECT;
	    refcnt = (refcnt * 10) + (*tp - '0');
	    tp++;
	 }

	 strncpy (item, (char *)op, IDTSIZE);
	 op += IDTSIZE;
	 item[IDTSIZE] = '\0';
#ifdef DEBUGLOADER
	 printf ("   IDT = %s", item);
#endif
	 tp = item;
	 for (; *tp; tp++) if (isspace(*tp)) *tp = '\0';
	 strcpy (module, item);
	 reloprog = *op++ == '1' ? TRUE : FALSE;
	 wdata = getnumfield (op, FALSE);
	 pseglen = wdata;
#ifdef DEBUGLOADER
	 printf (", length = %04X", pseglen);
#endif
	 if (pass == 1)
	 {
	    strcpy (modules[modcount].objfile, file);
	    strcpy (modules[modcount].name, item);
	    modules[modcount].length = wdata;
	    modules[modcount].origin = loadpt;
	    modules[modcount].creator[0] = '\0';
	    modules[modcount].date[0] = '\0';
	    modules[modcount].time[0] = '\0';
	    curmod = modcount;
	    modcount++;
	 }
	 else
	 {
	    curmod++;
	 }
#ifdef DEBUGLOADER
	 printf (", module = %s, curmod = %d", module, curmod);
#endif
	 if (idtbuf[0] == '\0')
	    strcpy (idtbuf, item);
      }

      /*
      ** Process DEF records
      */

      else if (defcnt)
      {
#ifdef DEBUGLOADER
	 printf ("   DEF: \n");
#endif
	 while (defcnt)
	 {
	    strncpy (item, (char *)op, MAXSYMLEN);
	    item[MAXSYMLEN] = '\0';
	    op += MAXSYMLEN;
	    wdata = getnumfield (op, FALSE);
	    op += WORDLEN;
	    if (reloprog)
	       wdata += loadpt;
#ifdef DEBUGLOADER
	    printf ("      sym = %s, val = %04X", item, wdata);
#endif
	    processdef (item, module, pass, reloprog, 0, wdata, libload);
#ifdef DEBUGLOADER
	    printf ("\n");
#endif
	    defcnt--;
	    if ((op-inbuf) >= 70)
	       break;
	 }
      }

      /*
      ** Process REF records
      */

      else if (refcnt)
      {
#ifdef DEBUGLOADER
	 printf ("   REF: \n");
#endif
	 while (refcnt)
	 {
	    strncpy (item, (char *)op, MAXSYMLEN);
	    item[MAXSYMLEN] = '\0';
	    op += MAXSYMLEN;
#ifdef DEBUGLOADER
	    printf ("      sym = %s", item);
#endif
	    processref (item, module, pass, reloprog, 0, 0);
#ifdef DEBUGLOADER
	    printf ("\n");
#endif
	    refcnt--;
	    if ((op-inbuf) >= 72)
	       break;
	 }
      }

      /*
      ** Process loadblocks
      */

      else
      {
	 int i, j;

#ifdef DEBUGLOADER
	 printf ("   LOAD BLOCK: loadpt = %04X, curraddr = %04X",
		 loadpt, curraddr);
#endif

	 if (!strncmp((char *)op, "/*", 2))
	 {
	    if (pass == 1)
	    {
	       if (reclen > 60)
	       {
		  strncpy (modules[curmod].date, (char *)&inbuf[TIMEOFFSET], 8);
		  modules[curmod].date[9] = '\0';
		  strncpy (modules[curmod].time, (char *)&inbuf[DATEOFFSET], 8);
		  modules[curmod].time[9] = '\0';
		  strncpy (modules[curmod].creator,
			   (char *)&inbuf[CREATOROFFSET], 8);
		  modules[curmod].creator[9] = '\0';
	       }
	       else
	       {
		  strncpy (modules[curmod].date, "        ", 8);
		  strncpy (modules[curmod].time, "        ", 8);
		  strncpy (modules[curmod].creator, "        ", 8);
	       }
	    }
	    if (curraddr < pseglen)
	    {
	       wdata = pseglen;
	       if (reloprog)
	          wdata += loadpt;
#ifdef DEBUGLOADER
	       printf (", ORG = %04X", wdata);
#endif
	       memctl[curraddr].tag = reloprog ? RELORG_TAG : ABSORG_TAG;
	       PUTMEM (curraddr, wdata);
	    }
	    curraddr = pseglen + loadpt;
	    loadpt = curraddr;
	    if (libload)
	       break;
	    idtrec = TRUE;
	    continue;
	 }

	 if (!strncmp((char *)op, "&&&&", 4))
	 {
	    op += WORDLEN;
	    wdata = getnumfield (op, FALSE);
#ifdef DEBUGLOADER
	    printf (", ENTRY = %04X\n", wdata);
#endif
	    if (reloprog)
	       relentry = (wdata + loadpt) & 0xFFFF;
	    else
	       absentry = wdata & 0xFFFF;
	    continue;
	 }

         wdata = getnumfield (op, FALSE);
	 if (reloprog)
	    wdata += loadpt;
	 op += WORDLEN;
	 if (wdata > curraddr)
	 {
#ifdef DEBUGLOADER
	    printf (", ORG = %04X", wdata);
#endif
	    memctl[curraddr].tag = reloprog ? RELORG_TAG : ABSORG_TAG;
	    PUTMEM (curraddr, wdata);
	 }
	 curraddr = wdata;

	 for (i = 0; i < 2; i++)
	 {
	    reloword = getnumfield (op, FALSE);
	    op += WORDLEN;
	    for (j = 0; j < 8; j++)
	    {
	       int relo;

	       if (*op == ' ')
	       {
	          i = j = 9;
		  continue;
	       }
	       wdata = getnumfield (op, FALSE);
	       op += WORDLEN;
	       relo = (reloword >> (14 - (j*2))) & 3;
	       switch (relo)
	       {
	       case 0: /* Absolute */
		  PUTMEM (curraddr, wdata);
		  memctl[curraddr].flags = 0;
		  memctl[curraddr].tag = ABSDATA_TAG;
		  curraddr += 1;
		  break;

	       case 2: /* Relocatable */
	          wdata += loadpt;
		  PUTMEM (curraddr, wdata);
		  memctl[curraddr].flags = RELOCATABLE;
		  memctl[curraddr].tag = RELDATA_TAG;
		  curraddr += 1;
		  break;

	       case 3: /* External */
		  PUTMEM (curraddr, wdata);
		  memctl[curraddr].flags = reloprog ? RELOCATABLE : 0;
		  memctl[curraddr].tag = EXTVALUE_TAG;
		  memctl[curraddr].value = 0;
		  if (pass == 2)
		     processextndx (curraddr, FALSE);
		  curraddr += 1;
	          break;

	       default:
	          fprintf (stderr, "lnk980: Unsupported relocation %d\n", relo);
		  exit (ABORT);
	       }
	    }
	 }
      }
#ifdef DEBUGLOADER
      printf ("\n");
#endif
   }
   return (curraddr);

BAD_OBJECT:
   fprintf (stderr, "lnk980: Bad object file\n");
   exit (ABORT);
}

/************************************************************************
* lnkloader - Object loader for lnk980.
************************************************************************/

int
lnkloader (FILE *fd, int loadpt, int pass, char *file, int libload)
{
   int binarymode = FALSE;
   int reclen;
   int curraddr;
   int relo;
   int done;
   int i;
   int relbyte;
   int wordlen = WORDTAGLEN-1;
   uint16 pseglen;
   unsigned char inbuf[82];
   char module[MAXSYMLEN+2];

   extndx = 0;
   for (i = 0; i < symbolcount; i++)
   {
      symbols[i]->extndx = -1;
   }

   i = fgetc (fd);
   ungetc (i, fd);
   if (i != IDT_TAG)
   {
      cardinmode = TRUE;
      return (loadcardfile (fd, loadpt, pass, file, libload));
   }

   curraddr = loadpt;
   cardinmode = FALSE;

#ifdef DEBUGLOADER
   printf ("lnkloader: loadpt = %04X, pass = %d, file = %s\n",
	   loadpt, pass, file);
#endif


   done = FALSE;
   pseglen = 0;
   while (!done)
   {
      unsigned char *op = inbuf;

      if (fgets ((char *)inbuf, 82, fd) == NULL)
      {
	 done = TRUE;
	 break;
      }
      reclen = strlen ((char *)inbuf);

#ifdef DEBUGLOADER
      printf ("reclen = %d\n", reclen);
#endif
      if (*op == EOFSYM)
      {
	 if (pass == 1)
	 {
	    if (reclen > 60)
	    {
	       strncpy (modules[curmod].date, (char *)&inbuf[TIMEOFFSET], 8);
	       modules[curmod].date[9] = '\0';
	       strncpy (modules[curmod].time, (char *)&inbuf[DATEOFFSET], 8);
	       modules[curmod].time[9] = '\0';
	       strncpy (modules[curmod].creator,
		        (char *)&inbuf[CREATOROFFSET], 8);
	       modules[curmod].creator[9] = '\0';
	    }
	    else
	    {
	       strncpy (modules[curmod].date, "        ", 8);
	       strncpy (modules[curmod].time, "        ", 8);
	       strncpy (modules[curmod].creator, "        ", 8);
	    }
	 }
	 loadpt = curraddr;
	 if (libload)
	    break;
	 continue;
      }

      relbyte = FALSE;
      for (i = 0; i < reclen; i++)
      {
	 char *bp;
	 char otag;
	 char item[80];
	 uint16 wdata;
	 uint16 bit;

	 relo = FALSE;
	 otag = *op++;

	 wdata = getnumfield (op, binarymode);

#ifdef DEBUGLOADER
	 printf ("   otag = %c, loadpt = %04X, curraddr = %04X",
		  isprint(otag) ? otag : '0', loadpt, curraddr);
#endif

         switch (otag)
	 {
	 case IDT_TAG:
	    op += wordlen;
	    strncpy (item, (char *)op, IDTSIZE);
	    item[IDTSIZE] = '\0';
#ifdef DEBUGLOADER
            printf (", IDT = %s", item);
#endif
	    bp = item;
	    for (; *bp; bp++) if (isspace(*bp)) *bp = '\0';
	    strcpy (module, item);
	    pseglen = wdata;
	    if (pass == 1)
	    {
	       strcpy (modules[modcount].objfile, file);
	       strcpy (modules[modcount].name, item);
	       modules[modcount].length = wdata;
	       modules[modcount].origin = loadpt;
	       modules[modcount].creator[0] = '\0';
	       modules[modcount].date[0] = '\0';
	       modules[modcount].time[0] = '\0';
	       curmod = modcount;
	       modcount++;
	    }
	    else
	    {
	       curmod++;
	    }
#ifdef DEBUGLOADER
            printf (", module = %s, curmod = %d", module, curmod);
#endif
	    if (idtbuf[0] == '\0')
	       strcpy (idtbuf, item);
	    op += IDTSIZE;
	    break;

         case RELORG_TAG:
	    wdata += loadpt;
         case ABSORG_TAG:
	    if (pseglen)
	    {
	       if (wdata >= curraddr)
	       {
#ifdef DEBUGLOADER
		  printf (", ORGTAG = %04X", wdata);
#endif
		  PUTMEM (curraddr, wdata);
		  memctl[curraddr].tag = otag;
		  curraddr = wdata & 0xFFFF;
	       }
#ifdef DEBUGLOADER
	       else
		  printf (", BACK ORGTAG = %04X", wdata);
#endif
	    }
	    op += wordlen;
	    break;

	 case RELDATA_TAG:
	    wdata += loadpt;
	    relo = TRUE;
	 case ABSDATA_TAG:
	    if (relbyte)
	    {
	       relbyte = FALSE;
	       bit = wdata & 0x0001;
	       wdata = (GETMEM (curraddr - 1) << 14) | (wdata >> 1);
	       wdata += loadpt;
	       PUTMEM (curraddr - 1, wdata >> 14);
	       wdata = ((wdata << 1) & 0x7FFE) | bit;
	    }
	    PUTMEM (curraddr, wdata);
	    memctl[curraddr].flags = relo ? RELOCATABLE : 0;
	    memctl[curraddr].tag = otag;
	    curraddr++;
	    op += wordlen;
	    break;

	 case RELENTRY_TAG:
	    relentry = (wdata + loadpt) & 0xFFFF;
	    op += wordlen;
	    break;

	 case ABSENTRY_TAG:
	    absentry = wdata & 0xFFFF;
	    op += wordlen;
	    break;

	 case RELEXTRN_TAG:
	    wdata += loadpt;
	    relo = TRUE;
	 case ABSEXTRN_TAG:
	    op += wordlen;
	    strncpy (item, (char *)op, MAXSYMLEN);
	    item[MAXSYMLEN] = '\0';
	    processref (item, module, pass, relo, 0, wdata);
	    op += MAXSYMLEN;
	    break;

	 case RELGLOBAL_TAG:
	    wdata += loadpt;
	    relo = TRUE;
	 case ABSGLOBAL_TAG:
	    op += wordlen;
	    strncpy (item, (char *)op, MAXSYMLEN);
	    item[MAXSYMLEN] = '\0';
	    processdef (item, module, pass, relo, 0, wdata, libload);
	    op += MAXSYMLEN;
	    break;

	 case RELBYTE_TAG:
	    relbyte = TRUE;
	    PUTMEM (curraddr, wdata);
	    memctl[curraddr].flags = relo ? RELOCATABLE : 0;
	    memctl[curraddr].tag = otag;
	    curraddr++;
	    op += wordlen;
	    break;

	 case EOR_TAG:
	    i = 81;
	    break;

	 case NOCKSUM_TAG:
	 case CKSUM_TAG:
	    op += wordlen;
	    break; 

	 default:
	    fprintf (stderr, "lnk980: UNSUPPORTED TAG, tag = %c, %s\n",
		     otag, &inbuf[71]);
	    exit (ABORT);
	 }
#ifdef DEBUGLOADER
	 printf ("\n");
#endif
      }
   }

   return (curraddr);

}
